unit KOLFlatTabControl;

//  file: KOLFlatTabControl.pas
//  file version: 0.1
//  last modified: 12.03.06
//  package: FlatControls
//  author: Karpinskyj Alexandr aka homm
//      mailto: homm86@mail.ru
//      My humble Web-Page: http://www.homm86.narod.ru

interface

uses
    Windows, Messages, KOL;

type
    ARect = array [0..4] of TRect;
    PARect = ^ARect;

    PFlatTabControl = ^TFlatTabControl;
    TKOLFlatTabControl = PFlatTabControl;

    TFlatTabControlPosition = (tpTop, tpBottom);

    TOnPageChanging = procedure(Sender: PFlatTabControl; NewPageIndex: Integer; var AllowChange: Boolean) of object;

    { TFlatTabControl }

    TFlatTabControl = object(TControl)
    private
        function GetFTC_Position: TFlatTabControlPosition;
        function GetFTC_TabsHeight: Integer;
        function GetTabsFont: PGraphicTool;
        procedure SetFTC_Position (AValue: TFlatTabControlPosition ) ;
        procedure SetFTC_TabsHeight ( const AValue: Integer ) ;
        function GetFTC_Items(Idx: Integer): String;
        procedure SetFTC_Items(Idx: Integer; const Value: String);
        function GetFTC_Images(Idx: Integer): Integer;
        procedure SetFTC_Images(Idx: Integer; const Value: Integer);
        function GetFTC_ItemRect(Idx: Integer): TRect;
    protected
        procedure DoPaint (Sender: PControl; DC:HDC);
        procedure DoMouseDown(const Msg: TMsg);
        procedure FTC_GetTabRects (var Rects: PARect);
    public
        function FTC_Insert ( Idx: Integer; const TabText: String; ImageIndex: Integer): PControl;
        procedure FTC_Delete ( Idx: Integer );
        property FTC_Items[ Idx: Integer ]: String read GetFTC_Items write SetFTC_Items;
        property FTC_Images[ Idx: Integer ]: Integer read GetFTC_Images write SetFTC_Images;
        property FTC_ItemRect[ Idx: Integer ]: TRect read GetFTC_ItemRect;
        function FTC_TabAtPos( x, y: Integer ): Integer;
        property FTC_Position: TFlatTabControlPosition read GetFTC_Position write SetFTC_Position;
        property FTC_TabsFont: PGraphicTool read GetTabsFont;
        property FTC_TabsHeight: Integer read GetFTC_TabsHeight write SetFTC_TabsHeight;
    end;

function NewFlatTabControl(AParent: PControl; const Tabs: array of PChar;
    ImgList: PImageList; ImgList1stIdx: Integer): PFlatTabControl;

const
    FTC_SetCurrent = WM_USER + $ceda;

implementation

type
    PFlatControlData = ^TFlatControlData;

    { TFlatControlData }

    TFlatControlData = object (TObj)
        FPosition: TFlatTabControlPosition;
        FOnPageChanging: TOnPageChanging;
        FTabsFont: PGraphicTool;
        FHeaderHeight: Integer;
        
        destructor Destroy; virtual;
    end;

{ TFlatControlData }

destructor TFlatControlData.Destroy;
begin
    FTabsFont.Free;
    inherited;
end;


function WndProcFlatTabControl ( Self_: PControl; var Msg: TMsg; var Rslt: Integer ): Boolean;
var     I, J: Integer;
        Page: PControl;
        Data: PFlatControlData;
        Allow: Boolean;
        Item: ^TTCItem;
begin
    Result := FALSE;
    Data := PFlatControlData(Self_.CustomObj);
    case Msg.message of
    WM_SIZE:
        begin
            for I := 0 to Self_.ChildCount-1 do begin
                Self_.Children[I].BoundsRect := Self_.ClientRect;
            end;
            Self_.Invalidate;
        end;
    WM_LBUTTONDOWN:
        begin
            PFlatTabControl(Self_).DoMouseDown(Msg);
        end;
    TCM_GETITEMCOUNT:
        begin
            Result := TRUE;
            Rslt := Self_.ChildCount;
        end;
    TCM_GETITEM:
        begin
            if Msg.wParam < Self_.ChildCount then begin
                Item := pointer(Msg.lParam);
                Item^.lParam := DWORD(Self_.Children[Msg.wParam]);
                Rslt := 1;
            end else
                Rslt := 0;
            Result := TRUE;
        end;
    FTC_SetCurrent:
        begin
            Rslt := 0;
            Result := TRUE;
            Allow := TRUE;
            I := Msg.wParam;
            if DWORD(I) >= DWORD(Self_.ChildCount) then exit;
            if Assigned(Data.FOnPageChanging) then
                Data.FOnPageChanging(PFlatTabControl(Self_), I, Allow);
            if not Allow then exit;
            PFlatTabControl(Self_).fCurIndex := I;
            for J := 0 to Self_.ChildCount-1 do begin
                Page := Self_.Children[ J ];
                Page.Visible := J = I;
                if J = I then
                    Page.BringToFront;
            end;
            Self_.Invalidate;
        end;
    end;
end;

function TFlatTabControl.FTC_Insert( Idx: Integer; const TabText: String; ImageIndex: Integer): PControl;
begin
    Result := NewPanel(@Self, esNone);
    Result.BoundsRect := Self.ClientRect;
    PFlatTabControl(Result).fREUrl := TabText;
    PFlatTabControl(Result).fCurIndex := ImageIndex;
    with Self.fChildren^ do
        MoveItem(Count-1, Idx);
    CurIndex := Idx;
    Invalidate;
    //////debug
    //Result.Caption := TabText;
    //Result.Color := clGreen;
end;

procedure TFlatTabControl.FTC_GetTabRects(var Rects: PARect);
var     ChCo: Integer;
        additR: PARect;

    function Summa (Ptr: PARect; Col: Integer): Integer;
    var     I: Integer;
    begin
        Result := 0;
        for I := 0 to Col-1 do
            inc(Result, Ptr[I].Right);
    end;

    function haveSmall(Rects: PARect; ChCo, Middle: Integer): Boolean;
    var     I: Integer;
    begin
        Result := FALSE;
        for I := 0 to ChCo-1 do
            if (Rects[I].Right < Middle) and (Rects[I].Right > 0) then begin
                Result := TRUE;
                exit;
            end;
    end;

    function smallest(Rects: PARect; ChCo: Integer): Integer;
    var     I: Integer;
    begin
        Result := 0;
        for I := 1 to ChCo-1 do begin
            if Rects[Result].Right = 0 then
                Result := I;
            if (Rects[I].Right > 0) and (Rects[I].Right < Rects[Result].Right) then
                Result := I;
        end;
    end;

    procedure GetTabRects(var Ptr: PARect; ChCo: Integer; Data: PFlatControlData);
    var     I: Integer;
            ACanvas: PCanvas;
    begin
        ACanvas := NewCanvas(0);
        ACanvas.Font.Assign(Data.FTabsFont);
        GetMem(Ptr, ChCo*16);
        FillChar(Ptr^, ChCo*16, 0);
        for I := 0 to ChCo-1 do begin
            Ptr[I].Right := ACanvas.TextWidth(Children[I].RE_URL) + 6;
            if Assigned(ImageListNormal) and (Children[I].CurIndex >= 0) then
                inc(Ptr[I].Right, ImageListNormal.ImgWidth + 4);
        end;
        ACanvas.Free;
    end;

    procedure MergeEx(DST, SRC: PARect; Middle: Integer{; ChCo: Integer});
    var     I: Integer;
    begin
        for I := 0 to ChCo-1 do begin
            if DST[I].Right = 0 then
                DST[I].Right := SRC[I].Right
            else
                DST[I].Right := Middle;
        end;
    end;

    procedure Expand(Rects: PARect; ChCo: Integer; Position: TFlatTabControlPosition);
    var     I: Integer;
            ptr: Integer;
            T: Integer;
    begin
        ptr := 4;
        for I := 0 to ChCo-1 do begin
            T := Rects[I].Right;
            OffsetRect(Rects[I], ptr, 0);
            inc(ptr, T);
            T := Height - ClientHeight;
            if Position = tpTop then begin
                Rects[I].Top := 3;
                Rects[I].Bottom := T;
            end else begin
                Rects[I].Top := 0;
                Rects[I].Bottom := T - 4;
            end;
        end;
    end;

var     I: Integer;
        Data: PFlatControlData;
        Aval: Integer;
        Middle: Integer;
        del: Integer;
begin
    Data := PFlatControlData(CustomObj);
    ChCo := ChildCount;
    GetTabRects(Rects, ChCo, Data);
    Aval := Width - 8;
    if Summa(Rects, ChCo) > Aval then begin
        Middle := Aval div ChCo;
        Del := 0;
        while haveSmall(Rects, ChCo, Middle) do begin
            I := Smallest(Rects, ChCo);
            dec(Aval, Rects[I].Right);
            Rects[I].Right := 0;
            inc(Del);
            Middle := Aval div (ChCo-del);
        end;
        GetTabRects(additR, ChCo, Data);
        MergeEx(Rects, additR, Aval div (ChCo-Del));
        FreeMem(additR);
    end;
    Expand(Rects, ChCo, Data^.FPosition);
end;


function TFlatTabControl.GetFTC_Position: TFlatTabControlPosition;
begin
    Result := PFlatControlData(CustomObj)^.FPosition;
end;

function TFlatTabControl.GetFTC_TabsHeight: Integer;
begin
    Result := PFlatControlData(CustomObj)^.FHeaderHeight;
end;

function TFlatTabControl.GetTabsFont: PGraphicTool;
begin
    Result := PFlatControlData(CustomObj)^.FTabsFont;
end;

procedure TFlatTabControl.SetFTC_Position ( AValue: TFlatTabControlPosition ) ;
var     Data: PFlatControlData;
begin
    Data := PFlatControlData(CustomObj);
    Data.FPosition := AValue;
    if AValue = tpTop then begin
        MarginTop := Data.FHeaderHeight;
        MarginBottom := 0;
    end else begin
        MarginBottom := Data.FHeaderHeight;
        MarginTop := 0;
    end;
    Perform(WM_SIZE, 0, 0);
end;

procedure TFlatTabControl.SetFTC_TabsHeight ( const AValue: Integer ) ;
begin
    PFlatControlData(CustomObj)^.FHeaderHeight := AValue;
    FTC_Position := FTC_Position;
end;

procedure TFlatTabControl.DoPaint (Sender: PControl; DC:HDC);
var     Bitmap: PBitmap;
        I: Integer;
        Data: PFlatControlData;
        T1: Integer;
        TR: TRect;
        Rects: PARect;
begin
    Data := PFlatControlData(CustomObj);
    Canvas.Brush.BrushStyle := bsSolid;
    Canvas.Brush.Color := Color;
    if not fTransparent then
        Canvas.FillRect(Self.ClientRect);
    Bitmap := NewBitmap(Width, MarginTop + MarginBottom);
    with Bitmap.Canvas^ do begin
        Font.Assign(Data.FTabsFont);
        Brush.Color := clBtnShadow;
        FillRect(Bitmap.BoundsRect);
        Pen.Color := clWindowFrame;
        Pen.PenStyle := psSolid;
        Pen.PenWidth := 1;
        if Data.FPosition = tpTop then begin
            T1 := Bitmap.Height-1;
        end else begin
            T1 := 0;
        end;
        MoveTo(0, T1);
        LineTo(Width, T1);

        FTC_GetTabRects(Rects);
        for I := 0 to ChildCount-1 do begin
            Pen.Color := clBtnHighlight;
            Brush.BrushStyle := bsSolid;
            if Self.CurIndex = I then begin
                Brush.Color := Self.Color;
                FillRect(Rects[I]);
                MoveTo(Rects[I].Left, Rects[I].Top);
                LineTo(Rects[I].Left, Rects[I].Bottom);
                if Data^.FPosition = tpTop then begin
                    MoveTo(Rects[I].Left, Rects[I].Top);
                    LineTo(Rects[I].Right, Rects[I].Top);
                end;
                Pen.Color := clWindowFrame;
                LineTo(Rects[I].Right, Rects[I].Bottom);
                LineTo(Rects[I].Right, Rects[I].Top);

                Font.Color := clWindowFrame;
            end else begin
                MoveTo(Rects[I].Right, 4);
                LineTo(Rects[I].Right, Bitmap.Height-3);
                Font.Color := clBtnHighlight;
            end;

            TR := Rects[I];
            InflateRect(TR, -3, 0);
            if (ImageListNormal <> nil) and (Children[I].CurIndex >= 0)
                and (TR.Right - TR.Left > ImageListNormal.ImgWidth) then begin
                TR.Left := TR.Left + ImageListNormal.ImgWidth + 4;
                T1 := (Rects[I].Bottom - Rects[I].Top - ImageListNormal.ImgHeight) div 2;

                ImageListNormal.Draw(Children[I].CurIndex, Handle, Rects[I].Left+4, Rects[I].Top + T1);
            end;

            Brush.BrushStyle := bsClear;
            DrawText(Children[I].RE_URL, TR, DT_VCENTER or DT_SINGLELINE or DT_END_ELLIPSIS);
        end;
        FreeMem(Rects);
    end;
    if Data^.FPosition = tpTop then
        T1 := 0
    else
        T1 := Height - Bitmap.Height;
    Bitmap.Draw(DC, 0, T1);
    Bitmap.Free;
end;

procedure TFlatTabControl.DoMouseDown ( const Msg: TMsg ) ;
var     Rects: PARect;
        I: Integer;
        TP: TPoint;
begin
    FTC_GetTabRects(Rects);

    TP := MakePoint(loword(Msg.lParam), hiword(Msg.lParam));
    if PFlatControlData(CustomObj).FPosition = tpBottom then
        dec(TP.Y, ClientHeight);
    for I := 0 to ChildCount-1 do
        if PtInRect(Rects[I], TP) then
            CurIndex := I;
    FreeMem(Rects);
end;

function NewFlatTabControl(AParent: PControl; const Tabs: array of PChar;
    ImgList: PImageList; ImgList1stIdx: Integer): PFlatTabControl;
var     I: Integer;
        Data: PFlatControlData;
        II: Integer;
begin
    Result := PFlatTabControl(_NewControl(AParent, 'FlatTabControl', WS_CHILD
        or WS_CLIPSIBLINGS or WS_CLIPCHILDREN or WS_VISIBLE, FALSE, @LabelActions));
    Result.fCommandActions.aSetCurrent := FTC_SetCurrent;
    Result.fCommandActions.aGetCount := TCM_GETITEMCOUNT;
    Result.Style := Result.Style;
    New(Data, Create);
    Data.FTabsFont := NewFont;
    Data.FTabsFont.FontName := 'Verdana';
    Data.FTabsFont.FontHeight := 13;
    Data.FTabsFont.FontStyle := [fsBold];
    Data.FHeaderHeight := 22;
    Result.CustomObj := Data;

    if ImgList <> nil then begin
        Result.ImageListNormal := ImgList;
        II := ImgList1stIdx;
        if ImgList.ImgHeight > 16 then
            Data.FHeaderHeight := ImgList.ImgHeight + 6;
    end else
        II := -1;
    for I := 0 to High(Tabs) do begin
        Result.FTC_Insert(I, Tabs[I], II);
        if ImgList <> nil then
            inc(II);
    end;
    Result.FTC_Position := tpBottom;
    Result.OnPaint := Result.DoPaint;
    Result.AttachProc (WndProcFlatTabControl);
    Result.CurIndex := 0;
end;

procedure TFlatTabControl.FTC_Delete(Idx: Integer);
begin
    Children[idx].Free;
end;

function TFlatTabControl.GetFTC_Items(Idx: Integer): String;
begin
    Result := Children[idx].RE_URL;
end;

procedure TFlatTabControl.SetFTC_Items(Idx: Integer; const Value: String);
begin
    PFlatTabControl(Children[idx]).fREUrl := Value;
end;

function TFlatTabControl.GetFTC_Images(Idx: Integer): Integer;
begin
    Result := Children[Idx].CurIndex;
end;

procedure TFlatTabControl.SetFTC_Images(Idx: Integer;
  const Value: Integer);
begin
    Children[Idx].CurIndex := Value;
end;

function TFlatTabControl.GetFTC_ItemRect(Idx: Integer): TRect;
var     TRects: PARect;
begin
    FTC_GetTabRects(TRects);
    Result := TRects^[Idx];
    if FTC_Position = tpBottom then
        OffsetRect(Result, 0, Height - Self.MarginTop - Self.MarginBottom);
    FreeMem(TRects);
end;

function TFlatTabControl.FTC_TabAtPos(x, y: Integer): Integer;
var     TP: TPoint;
        TRects: PARect;
        I: Integer;
        orTop: Boolean;
begin
    TP := MakePoint(X, Y);
    FTC_GetTabRects(TRects);
    orTop := FTC_Position = tpTop;
    Result := -1;
    for I := 0 to ChildCount do begin
        if not orTop then
            OffsetRect(TRects^[I], 0, Height - Self.MarginTop - Self.MarginBottom);
        if PtInRect(TRects^[I], TP) then
            Result := I;
    end;
    FreeMem(TRects);
end;

end.
